"""
pyaud._default
==============
"""
from . import _config
from . import plugins as _plugins
from ._locations import NAME as _NAME
from ._utils import colors as _colors
from ._utils import git as _git


class _Audit(_plugins.Action):
    """Read from [audit] key in config."""

    def action(self, *args: str, **kwargs: bool) -> int:
        funcs = _config.toml["audit"]["modules"]
        if kwargs.get("clean", False):
            funcs.insert(0, "clean")

        for func in funcs:
            if func in _plugins.registered():
                _colors.cyan.bold.print(f"\n{_NAME} {func}")
                _plugins.get(func)(**kwargs)

        return 0


class _GenerateRCFile(_plugins.Action):
    """Print rcfile to stdout.

    Print rcfile to stdout, so it may be piped to chosen filepath.
    """

    cache = False

    def action(self, *args: str, **kwargs: bool) -> int:
        print(_config.toml.dumps(_config.DEFAULT_CONFIG), end="")
        return 0


class _Clean(_plugins.Action):
    """Remove all unversioned package files recursively."""

    cache = False

    def action(self, *args: str, **kwargs: bool) -> int:
        exclude = _config.toml["clean"]["exclude"]
        return _git.clean(
            "-fdx", *[f"--exclude={e}" for e in exclude], **kwargs
        )


def register_default_plugins() -> None:
    """Register default plugins."""
    _plugins.register("audit")(_Audit)
    _plugins.register("generate-rcfile")(_GenerateRCFile)
    _plugins.register("clean")(_Clean)



"""
pyaud.exceptions
================

Exceptions for use within the module.

All exceptions made public for if they need to be reraised or excepted.

Exceptions are already built into the architecture but can be used in
new plugins as well.
"""
import typing as _t


class AuditError(Exception):
    """Raise for audit failures that aren't failed subprocesses.

    :param cmd: Command that failed. If no argument provided the value
        will be None.
    """

    def __init__(self, cmd: _t.Optional[str]) -> None:
        super().__init__(f"{cmd} did not pass all checks")


class NameConflictError(Exception):
    """Raise if adding plugin whose name is not unique.

    :param plugin: Plugin which could not be registered.
    :param name: Name which clashes with another.
    """

    def __init__(self, plugin: str, name: str) -> None:
        super().__init__(f"plugin name conflict at {plugin}: '{name}'")



"""
pyaud._wraps
============

Decorators for wrapping plugins.
"""
import functools as _functools
import sys as _sys
import typing as _t
import warnings as _warnings
from pathlib import Path as _Path

import spall.exceptions as sp_exceptions

from . import _data
from ._cache import FileCacher as _FileCacher
from ._indexing import files as _files
from ._locations import AppFiles as _AppFiles
from ._objects import BasePlugin as _BasePlugin
from ._utils import colors as _colors


class CheckCommand:
    """Decorate callable with status of completion."""

    @staticmethod
    def _announce_completion(success_message: str, returncode: int) -> None:
        if returncode:
            _colors.red.bold.print(
                f"Failed: returned non-zero exit status {returncode}",
                file=_sys.stderr,
            )
        else:
            _colors.green.bold.print(success_message)

    @classmethod
    def file(cls, func: _t.Callable[..., int]) -> _t.Callable[..., int]:
        """Run the routine common with single file fixes.

        :param func: Function to decorate.
        :return: Wrapped function.
        """

        @_functools.wraps(func)
        def _wrapper(*args: str, **kwargs: bool) -> int:
            returncode = func(*args, **kwargs)
            cls._announce_completion(
                "Success: no issues found in file", returncode
            )
            return returncode

        return _wrapper

    @classmethod
    def files(cls, func: _t.Callable[..., int]) -> _t.Callable[..., int]:
        """Run the routine common with multiple source file fixes.

        :param func: Function to decorate.
        :return: Wrapped function.
        """

        @_functools.wraps(func)
        def _wrapper(*args: str, **kwargs: bool) -> int:
            returncode = 0
            if not _files.reduce():
                print("No files found")
            else:
                returncode = func(*args, **kwargs)
                cls._announce_completion(
                    f"Success: no issues found in {len(_files)} source files",
                    returncode,
                )

            return returncode

        return _wrapper


class ClassDecorator:
    """Handle reading and writing file data for called processes.

    Decorate on call to ``__new__`` to wrap uninstantiated class and its
    ``__call__`` method.

    :param cls: The class whose ``__call__`` method will be wrapped.
    :param app_files: App file locations object.
    """

    def __init__(
        self, cls: _t.Type[_BasePlugin], app_files: _AppFiles
    ) -> None:
        self._cls = cls
        self._app_files = app_files

    def time(self, func: _t.Callable[..., int]) -> _t.Callable[..., int]:
        """Wrap ``__call__`` with a timer.

        :param func: Function to wrap.
        :return: Wrapped function.
        """

        @_functools.wraps(func)
        def _wrapper(*args: str, **kwargs: bool) -> int:
            repo = _Path.cwd().name
            with _data.record.track(
                repo, self._cls, self._app_files.durations_file
            ) as time_keeper:
                returncode = func(*args, **kwargs)

            _data.record.write(self._app_files.durations_file)
            logged_time = "{}: Execution time: {}s; Average time: {}s".format(
                self._cls.__name__,
                time_keeper.elapsed(),
                _data.record.average(repo, self._cls),
            )
            self._cls.logger().info(logged_time)
            if kwargs.get("timed", False):
                _colors.magenta.print(logged_time)

            return returncode

        return _wrapper

    def files(self, func: _t.Callable[..., int]) -> _t.Callable[..., int]:
        """Wrap ``__call__`` with a hashing function.

        :param func: Function to wrap.
        :return: Wrapped function.
        """

        @_functools.wraps(func)
        def _wrapper(*args: str, **kwargs: bool) -> int:
            _file_cacher = _FileCacher(
                self._cls, func, self._app_files, *args, **kwargs
            )
            return _file_cacher.files(func, *args, **kwargs)

        return _wrapper

    @staticmethod
    def not_found(func: _t.Callable[..., int]) -> _t.Callable[..., int]:
        """Wrap ``__call__`` to resolve ``CommandNotFound`` errors.

        :param func: Function to wrap.
        :return: Wrapped function.
        """

        @_functools.wraps(func)
        def _wrapper(*args: str, **kwargs: bool) -> int:
            try:
                return func(*args, **kwargs)
            except sp_exceptions.CommandNotFoundError as err:
                _warnings.warn(
                    f"{str(err).split(':', maxsplit=1)[0]}: Command not found",
                    RuntimeWarning,
                )

            return 1

        return _wrapper



"""
pyaud._typing
=============
"""
import typing as _t
from types import TracebackType as _TracebackType

Exc = _t.Optional[_t.Type[BaseException]]
ExcVal = _t.Optional[BaseException]
ExcTB = _t.Optional[_TracebackType]



"""Framework for writing Python package audits."""
from . import exceptions, plugins
from ._indexing import files
from ._objects import BasePlugin
from ._utils import package
from ._version import __version__

__all__ = [
    "BasePlugin",
    "__version__",
    "exceptions",
    "files",
    "package",
    "plugins",
]



"""
pyaud._version
==============

Contains the version of this package.

Allows for access to the version internally without cyclic imports
caused by accessing it through __init__.
"""
__version__ = "4.1.0"



"""
pyaud._data
===========

Persistent data for use between runs.
"""
from __future__ import annotations

import contextlib as _contextlib
import typing as _t
from pathlib import Path as _Path
from time import time as _time

from ._objects import JSONIO as _JSONIO
from ._objects import BasePlugin as _BasePlugin


class _TimeKeeper:
    def __init__(self, cls: _t.Type[_BasePlugin]) -> None:
        self._cls = cls
        self._start_time = 0.0
        self._end_time = self._start_time
        self._elapsed = self._start_time

    def _starter(self):
        self._start_time = _time()

    def _stopper(self):
        self._end_time = _time()

    def start(self) -> None:
        """Start the timer."""
        self._starter()
        self._end_time = self._start_time

    def stop(self) -> None:
        """Record the time from entry to exit."""
        self._stopper()
        self._elapsed = round(self._end_time - self._start_time, 2)

    def elapsed(self) -> float:
        """Return the elapsed time.

        :return: The elapsed time.
        """
        return self._elapsed


class Record(_JSONIO):
    """Record floats to objects based on calling class."""

    def average(self, repo: str, cls: _t.Type[_BasePlugin]) -> float:
        """Get the average of all recorded times.

        :param repo: Name of package audit is running in.
        :param cls: Name of class that this is running in.
        :return: Float containing the average of elapsed times.
        """
        items = self.get(repo, {}).get(str(cls), [])
        return round(sum(items) / len(items), 2)

    @_contextlib.contextmanager
    def track(
        self, repo: str, cls: _t.Type[_BasePlugin], path: _Path
    ) -> _t.Generator[_TimeKeeper, None, None]:
        """Context manager for parsing envvars with a common prefix.

        :param repo: Name of package audit is running in.
        :param cls: Name of class that this is running in.
        :param path: Path to datafile.
        :return: ``Instantiated _TimeKeeper`` object.
        """
        time_keeper = _TimeKeeper(cls)
        try:
            self[repo] = self.get(repo, {})
            self[repo][str(cls)] = self[repo].get(str(cls), [])
            time_keeper.start()
            yield time_keeper
        finally:
            time_keeper.stop()
            self[repo][str(cls)].append(time_keeper.elapsed())
            super().write(path)


record = Record()



"""
pyaud._objects
==============
"""
import json as _json
import logging as _logging
import typing as _t
from abc import ABC as _ABC
from pathlib import Path as _Path

_KT = _t.TypeVar("_KT")
_VT = _t.TypeVar("_VT")
_T_co = _t.TypeVar("_T_co", covariant=True)


class MutableMapping(_t.MutableMapping[_KT, _VT]):
    """Inherit to replicate subclassing of ``dict`` objects."""

    def __init__(self) -> None:
        self._dict: _t.Dict[_KT, _VT] = {}

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._dict}>"

    def __len__(self) -> int:
        return self._dict.__len__()

    def __delitem__(self, key: _KT) -> None:
        self._dict.__delitem__(key)

    def __setitem__(self, key: _KT, value: _VT) -> None:
        self._dict = self._nested_update(self._dict, {key: value})

    def __getitem__(self, key: _KT) -> _VT:
        return self._dict.__getitem__(key)

    def __iter__(self) -> _t.Iterator[_KT]:
        return self._dict.__iter__()

    def _nested_update(
        self, obj: _t.Dict[_KT, _t.Any], update: _t.Dict[_KT, _t.Any]
    ) -> _t.Dict[_KT, _t.Any]:
        # add to __setitem__ to ensure that no entire dict keys with
        # missing nested keys overwrite all other values
        # run recursively to cover all nested objects if value is a dict
        # if value is a str pass through ``Path.expanduser()`` to
        # translate paths prefixed with ``~/`` for ``/home/<user>``
        # if value is all else assign it to obj key
        # return obj for recursive assigning of nested dicts
        for key, value in update.items():
            if isinstance(value, dict):
                value = self._nested_update(obj.get(key, {}), value)

            elif isinstance(value, str):
                value = str(_Path(value).expanduser())

            obj[key] = value

        return obj


class BasePlugin(_ABC):  # pylint: disable=too-few-public-methods
    """Base type for all plugins."""

    #: If set to True then indexed files will be monitored for change.
    cache = False

    #: Only matters if ``cache`` is set to True.
    #: If False (default) then audit will cache on a file-by-file basis.
    #: If True, then no changes can be made to any file for a cache-hit
    #: to be valid.
    cache_all = False

    #: set a single cache file for plugin subclass.
    cache_file: _t.Optional[_t.Union[str, _Path]] = None

    @classmethod
    def logger(cls) -> _logging.Logger:
        """Assign an audit logger dynamically, post logging config.

        :return: ``Logger`` object.
        """
        return _logging.getLogger(cls.__name__)


class JSONIO(MutableMapping):
    """Base class JSON input/output actions."""

    def read(self, path: _Path) -> None:
        """Read from file to object.

        :param path: Path to json file.
        """
        if path.is_file():
            try:
                self.update(_json.loads(path.read_text()))
            except _json.decoder.JSONDecodeError:
                pass

    def write(self, path: _Path) -> None:
        """Write data to file.

        :param path: Path to json file.
        """
        path.write_text(_json.dumps(dict(self), separators=(",", ":")))



"""
pyaud.plugins
=============

Main module used for public API.
"""
from __future__ import annotations

import importlib as _importlib
import inspect as _inspect
import os as _os
import pkgutil as _pkgutil
import re
import sys as _sys
import typing as _t
from abc import abstractmethod as _abstractmethod
from pathlib import Path as _Path
from subprocess import CalledProcessError as _CalledProcessError

from spall import Subprocess as _Subprocess

from . import _config
from . import exceptions as _exceptions
from ._indexing import files as _files
from ._locations import NAME as _NAME
from ._locations import AppFiles as _AppFiles
from ._objects import BasePlugin as _BasePlugin
from ._objects import MutableMapping as _MutableMapping
from ._utils import colors as _colors
from ._wraps import CheckCommand as _CheckCommand
from ._wraps import ClassDecorator as _ClassDecorator


class _SubprocessFactory(_MutableMapping):
    """Instantiate collection of ``Subprocess`` objects."""

    def __init__(self, args: _t.List[str]) -> None:
        super().__init__()
        for arg in args:
            self[arg] = _Subprocess(arg)


class Plugin(_BasePlugin):
    """Base class of all plugins.

    Raises ``TypeError`` if registered directly.

    Contains the name attribute assigned upon registration.

    Subprocesses are stored in the ``subprocess`` dict object

    :param name: Name assigned to plugin via ``@register`` decorator.
    """

    def __new__(cls, name: str) -> Plugin:  # pylint: disable=unused-argument
        app_files = _AppFiles()
        class_decorator = _ClassDecorator(cls, app_files)
        cls.__call__ = class_decorator.not_found(cls.__call__)  # type: ignore
        cls.__call__ = class_decorator.files(cls.__call__)  # type: ignore
        cls.__call__ = class_decorator.time(cls.__call__)  # type: ignore
        return super().__new__(cls)

    def __init__(self, name: str) -> None:
        self.name = name
        self.subprocess = _SubprocessFactory(self.exe)

    def __deepcopy__(self, name: str) -> Plugin:
        return self

    @staticmethod
    def audit_error() -> _exceptions.AuditError:
        """Raise if checks have failed.

        :return: AuditError instantiated with error message.
        """
        return _exceptions.AuditError(" ".join(_sys.argv))

    @property
    def env(self) -> _t.Dict[str, str]:
        """Return environment which will remain active for run."""
        return {}

    @property
    def exe(self) -> _t.List[str]:
        """List of executables to add to ``subprocess`` dict."""
        return []

    def __call__(self, *args: str, **kwargs: bool) -> int:
        """Enables calling of all plugin instances."""
        return 0


class Audit(Plugin):
    """Blueprint for writing audit-only plugins.

    Audit will be called from here.

    Run within context of defined environment variables.
    If no environment variables are defined nothing will change.

    :raises CalledProcessError: Will always be raised if something fails
        that is not to do with the audit condition. Will be excepted and
        reraised as ``AuditError`` if the audit fails.
    :raises AuditError: Raised from ``CalledProcessError`` if
        audit fails.
    :return: If any error has not been raised for any reason int object
        must be returned, from subprocess or written, to notify call
        whether process has succeeded or failed. No value will actually
        return from __call__ as it will be passed to the decorator.
    """

    @_abstractmethod
    def audit(self, *args: str, **kwargs: bool) -> int:
        """All audit logic to be written within this method.

        :param args: Args that can be passed from other plugins.
        :param kwargs: Boolean flags for subprocesses.
        :return: If any error has not been raised for any reason int
            object must be returned, from subprocess or written, to
            notify call whether process has succeeded or failed.
        """

    @_CheckCommand.files
    def __call__(self, *args: str, **kwargs: bool) -> int:
        with _config.TempEnvVar(_os.environ, **self.env):
            try:
                return self.audit(*args, **kwargs)

            except _CalledProcessError as err:
                raise self.audit_error() from err


#: Blueprint for writing audit and fix plugins.
class BaseFix(Audit):
    """Blueprint for writing audit and fix plugins.

    Audit will be called from here.

    Called within context of defined environment variables.
    If no environment variables are defined nothing will change.

    If audit fails and the ``-f/--fix`` flag is passed to the
    commandline the ``fix`` method will be called within the
    ``CalledProcessError`` try-except block.

    If ``-f/--fix`` and the audit fails the user is running the
    audit only and will raise an ``AuditError``.

    :raises CalledProcessError: Will always be raised if something fails
        that is not to do with the audit condition. Will be excepted and
        reraised as ``AuditError`` if the audit fails and ``-f/--fix``
        is not passed to the commandline.
    :raises AuditError: Raised from ``CalledProcessError`` if audit
        fails and ``-f/--fix`` flag if not passed to the commandline.
    :return: If any error has not been raised for any reason int object
        must be returned, from subprocess or written, to notify call
        whether process has succeeded or failed. No value will actually
        return from __call__ as it will be passed to the decorator.
    """

    @_abstractmethod
    def audit(self, *args: str, **kwargs: bool) -> int:
        """All audit logic to be written within this method.

        :param args: Args that can be passed from other plugins.
        :param kwargs: Boolean flags for subprocesses.
        :return: If any error has not been raised for any reason int
            object must be returned, from subprocess or written, to
            notify __call__ whether process has succeeded or failed. If
            non-zero exist is returned and ``-f/--fix`` has been passed
            to the commandline run the ``fix`` method, otherwise raise
            ``AuditError``.
        """

    @_abstractmethod
    def fix(self, *args: str, **kwargs: bool) -> int:
        """Run if audit fails but only if running a fix.

        :param args: Args that can be passed from other plugins.
        :param kwargs: Boolean flags for subprocesses.
        :return: If any error has not been raised for any reason int
            object must be returned, from subprocess or written, to
            notify __call__ whether process has succeeded or failed.
        """

    def __call__(self, *args: str, **kwargs: bool) -> int:
        with _config.TempEnvVar(_os.environ, **self.env):
            try:
                returncode = self.audit(*args, **kwargs)

            except _CalledProcessError:
                returncode = 1

        if returncode:
            if kwargs.get("fix", False):
                return self.fix(**kwargs)

            raise self.audit_error()

        return returncode


class Fix(BaseFix):
    """Blueprint for writing audit and fix plugins for single files.

    Announce file status.

    :raises CalledProcessError: Will always be raised if something fails
        that is not to do with the audit condition. Will be excepted and
        reraised as ``AuditError`` if the audit fails and ``-f/--fix``
        is not passed to the commandline.
    :raises AuditError: Raised from ``CalledProcessError`` if audit
        fails and ``-f/--fix`` flag if not passed to the commandline.
    :return: If any error has not been raised for any reason int object
        must be returned, from subprocess or written, to notify call
        whether process has succeeded or failed. No value will actually
        return from __call__ as it will be passed to the decorator.
    """

    @_abstractmethod
    def audit(self, *args: str, **kwargs: bool) -> int:
        """All audit logic to be written within this method.

        :param args: Args that can be passed from other plugins.
        :param kwargs: Boolean flags for subprocesses.
        :return: If any error has not been raised for any reason int
            object must be returned, from subprocess or written, to
            notify __call__ whether process has succeeded or failed. If
            non-zero exist is returned and ``-f/--fix`` has been passed
            to the commandline run the ``fix`` method, otherwise raise
            ``AuditError``.
        """

    @_abstractmethod
    def fix(self, *args: str, **kwargs: bool) -> int:
        """Run if audit fails but only if running a fix.

        :param args: Args that can be passed from other plugins.
        :param kwargs: Boolean flags for subprocesses.
        :return: If any error has not been raised for any reason int
            object must be returned, from subprocess or written, to
            notify __call__ whether process has succeeded or failed.
        """

    @_CheckCommand.file
    def __call__(self, *args: str, **kwargs: bool) -> int:
        return super().__call__(*args, **kwargs)


class FixAll(BaseFix):
    """Blueprint for writing audit and fix plugins for Python files.

    Announce Python file status.

    :raises CalledProcessError: Will always be raised if something fails
        that is not to do with the audit condition. Will be excepted and
        reraised as ``AuditError`` if the audit fails and ``-f/--fix``
        is not passed to the commandline.
    :raises AuditError: Raised from ``CalledProcessError`` if audit
        fails and ``-f/--fix`` flag if not passed to the commandline.
    :return: If any error has not been raised for any reason int object
        must be returned, from subprocess or written, to notify call
        whether process has succeeded or failed. No value will actually
        return from __call__ as it will be passed to the decorator.
    """

    @_abstractmethod
    def audit(self, *args: str, **kwargs: bool) -> int:
        """All audit logic to be written within this method.

        :param args: Args that can be passed from other plugins.
        :param kwargs: Boolean flags for subprocesses.
        :return: If any error has not been raised for any reason int
            object must be returned, from subprocess or written, to
            notify __call__ whether process has succeeded or failed. If
            non-zero exist is returned and ``-f/--fix`` has been passed
            to the commandline run the ``fix`` method, otherwise raise
            ``AuditError``.
        """

    @_abstractmethod
    def fix(self, *args: str, **kwargs: bool) -> int:
        """Run if audit fails but only if running a fix.

        :param args: Args that can be passed from other plugins.
        :param kwargs: Boolean flags for subprocesses.
        :return: If any error has not been raised for any reason int
            object must be returned, from subprocess or written, to
            notify __call__ whether process has succeeded or failed.
        """

    @_CheckCommand.files
    def __call__(self, *args: str, **kwargs: bool) -> int:
        return super().__call__(*args, **kwargs)


class Action(Plugin):
    """Blueprint for writing generic plugins.

    Called within context of defined environment variables.
    If no environment variables are defined nothing will change.

    :raises CalledProcessError: Will always be raised if something fails
        that is not to do with the action condition. Will be excepted
        and reraised as ``AuditError`` if the action fails.
    :raises AuditError: Raised from ``CalledProcessError`` if action
        fails.
    :return: Any value and type can be returned.
    """

    @_abstractmethod
    def action(self, *args: str, **kwargs: bool) -> int:
        """All logic to be written within this method.

        :param args: Args that can be passed from other plugins.
        :param kwargs: Boolean flags for subprocesses.
        :return: Any value and type can be returned.
        """

    def __call__(self, *args: str, **kwargs: bool) -> int:
        with _config.TempEnvVar(_os.environ, **self.env):
            try:
                return self.action(*args, **kwargs)

            except _CalledProcessError as err:
                raise self.audit_error() from err


class Parametrize(Plugin):
    """Define a list of strings to call multiple plugins.

    :raises CalledProcessError: Will always be raised if something fails
        that is not to do with the called plugin's condition. Will be
        excepted and reraised as ``AuditError`` if the called plugin
        fails  and the called plugin does not specify a ``fix`` method
        or the ``-f/--fix`` flag is not passed to the commandline.
    :raises AuditError: Raised from ``CalledProcessError`` if called
        plugin fails and no ``fix`` method is specified or the
        ``-f/--fix`` flag is not passed to the commandline.
    """

    @_abstractmethod
    def plugins(self) -> _t.List[str]:
        """List of plugin names to run.

        :return: List of plugin names, as defined in ``@register``.
        """

    def __call__(self, *args: str, **kwargs: bool) -> int:
        for name in self.plugins():
            _colors.cyan.bold.print(f"\n{_NAME} {name}")
            _plugins[name](*args, **kwargs)

        # all parametrized plugins will have succeeded to make it here
        return 0


class FixFile(Plugin):
    """Blueprint for writing audit and fix plugins for individual files.

    All logic can act on each file that would be passed from __call__.

    Called within context of defined environment variables.
    If no environment variables are defined nothing will change.

    Condition for failure needs to be defined, as the file argument
    passed from outer loop will not return an exit status.

    If audit fails and the ``-f/--fix`` flag is passed to the
    commandline the ``fix`` method will be called within the
    ``CalledProcessError`` try-except block.

    If ``-f/--fix`` and the audit fails the user is running the
    audit only and will raise an ``AuditError``.

    :raises CalledProcessError: Will always be raised if something fails
        that is not to do with the audit condition. Will be excepted and
        reraised as ``AuditError`` if the audit fails and ``-f/--fix``
        is not passed to the commandline.
    :raises AuditError: Raised from ``CalledProcessError`` if audit
        fails and ``-f/--fix`` flag if not passed to the commandline.
    :return: Only 0 exit-status can be returned. If process fails error
        will be raised.
    """

    @_abstractmethod
    def fail_condition(self) -> _t.Optional[bool]:
        """Condition to trigger non-subprocess failure.

        :return: Return True for passing condition, false for failing
            condition, or None for no fail condition.
        """

    @_abstractmethod
    def audit(self, file: _Path, **kwargs: bool) -> int:
        """All logic written within this method for each file's audit.

        :param file: Individual file.
        :param kwargs: Boolean flags for subprocesses.
        :return: Returncode.
        """

    @_abstractmethod
    def fix(self, file: _Path, **kwargs: bool) -> int:
        """All logic written within this method for each file's fix.

        :param file: Individual file.
        :param kwargs: Boolean flags for subprocesses.
        :return: Returncode.
        """

    @_CheckCommand.files
    def __call__(self, *args: str, **kwargs: bool) -> int:
        returncode = 0
        files = [p for p in _files if p.is_file()]
        for file in files:
            returncode = self.audit(file, **kwargs)
            fail = self.fail_condition()
            if fail is not None and fail:
                if kwargs.get("fix", False):
                    return self.fix(file, **kwargs)

                raise self.audit_error()

        # if no error raised return 0 to decorator
        return returncode


# array of plugins
PLUGINS = [Audit, BaseFix, Fix, FixAll, Action, Parametrize, FixFile]

# array of plugin names
PLUGIN_NAMES = [t.__name__ for t in PLUGINS]

# array of plugin types before instantiation
PluginType = _t.Union[
    _t.Type[Audit],
    _t.Type[BaseFix],
    _t.Type[Fix],
    _t.Type[FixAll],
    _t.Type[Action],
    _t.Type[Parametrize],
    _t.Type[FixFile],
]

# array of plugin types after instantiation
PluginInstance = _t.Union[
    Audit, BaseFix, Fix, FixAll, Action, Parametrize, FixFile
]


class Plugins(_MutableMapping):
    """Holds registered plugins.

    Instantiate plugin on running __setitem__.

    :raise NameConflictError: If name of registered plugin is not
        unique.
    :raise TypeError: If non plugin type registered.
    """

    def __setitem__(self, name: str, plugin: PluginType) -> None:
        # only unique names to be set in `plugins` object
        # if name is not unique raise `NameConflictError`
        if name in self:
            raise _exceptions.NameConflictError(plugin.__name__, name)

        mro = tuple(p.__name__ for p in _inspect.getmro(plugin))
        if not hasattr(plugin, "__bases__") or not any(
            i in PLUGIN_NAMES for i in mro
        ):
            raise TypeError(
                "can only register one of the following: {}; not {}".format(
                    ", ".join(PLUGIN_NAMES), mro
                )
            )

        super().__setitem__(name, plugin(name))


_plugins = Plugins()


def _name_plugin(plugin: PluginType) -> str:
    parts = re.findall("[A-Z][^A-Z]*", plugin.__name__)
    return "-".join(parts).lower()


def register(
    name: _t.Optional[str] = None,
) -> _t.Callable[[PluginType], PluginType]:
    """Register subclassed plugin to collection.

    If name is not provided a name will be assigned automatically.

    :param name: Name to register plugin as.
    :return: Return registered plugin to call.
    """

    def _register(plugin: PluginType) -> PluginType:
        _plugins[name or _name_plugin(plugin)] = plugin
        return plugin

    return _register


def mapping() -> _t.Dict[str, PluginInstance]:
    """Get dict of named keys and their corresponding plugin values.

    :return: Mapping of plugins and their unique names.
    """
    return dict(_plugins)


def registered() -> _t.List[str]:
    """Get list of registered plugins.

    :return: List of registered plugins.
    """
    return sorted(list(_plugins))


def get(name: str) -> PluginInstance:
    """Get plugins by name.

    :param name: Unique name of plugin.
    :return: Callable plugin instance.
    """
    return _plugins[name]


def load() -> None:
    """Import all package prefixed with ``pyaud[-_]``."""
    for _, name, _ in _pkgutil.iter_modules():
        if re.match("^pyaud[-_].*$", name):
            _importlib.import_module(name)



"""
pyaud._main
===========

Contains package entry point.
"""
import logging as _logging
from pathlib import Path as _Path

from . import _config, _data
from . import plugins as _plugins
from ._cli import Parser as _Parser
from ._default import register_default_plugins as _register_default_plugins
from ._indexing import files as _files
from ._locations import AppFiles as _AppFiles


def main() -> None:
    """Module entry point.

    Parse commandline arguments and run the selected choice from the
    dictionary of functions which matches the key.
    """
    app_files = _AppFiles()
    _register_default_plugins()
    _plugins.load()
    parser = _Parser()
    _data.record.read(app_files.durations_file)
    _config.load_config(app_files, parser.args.rcfile)
    _config.configure_logging(parser.args.verbose)
    _files.add_exclusions(*_config.toml["indexing"]["exclude"])
    _files.populate()
    _logging.getLogger(__name__).info(
        "commencing audit for %s in %s",
        app_files.user_project_dir.name,
        _Path.cwd(),
    )
    _plugins.get(parser.args.module)(
        clean=parser.args.clean,
        suppress=parser.args.suppress,
        fix=parser.args.fix,
        timed=parser.args.timed,
        no_cache=parser.args.no_cache,
    )



"""
pyaud._indexing
===============
"""
from __future__ import annotations

import typing as _t
from types import TracebackType as _TracebackType

from lsfiles import LSFiles as _LSFiles


class IndexedState:
    """Store index and ensure it's in its original state on exit."""

    def __init__(self) -> None:
        self.length = len(files)
        self._index = list(files)
        self._restored = False

    def __enter__(self) -> IndexedState:
        return self

    def __exit__(
        self,
        exc_type: _t.Optional[_t.Type[BaseException]],
        exc_val: _t.Optional[BaseException],
        exc_tb: _t.Optional[_TracebackType],
    ) -> None:
        if not self._restored:
            files.extend(self._index)

    def restore(self) -> None:
        """Restore the original state of index."""
        self._restored = True
        files.extend(self._index)


files = _LSFiles()



"""
pyaud._locations
================
"""
from __future__ import annotations

from pathlib import Path as _Path

from appdirs import AppDirs as _AppDirs

NAME = __name__.split(".", maxsplit=1)[0]


class AppDirs(_AppDirs):
    """Directories for app to interact with.

    Create app's data dirs on instantiation.
    """

    def __init__(self) -> None:
        super().__init__(appname=NAME)
        for path in (
            self.user_config_dir,
            self.user_data_dir,
            self.user_cache_dir,
        ):
            path.mkdir(exist_ok=True, parents=True)

    @property
    def user_home_dir(self) -> _Path:
        """Path to user's home directory."""
        return _Path.home()

    @property
    def user_project_dir(self) -> _Path:
        """path to user's current project."""
        return _Path.cwd()

    @property
    def user_config_dir(self) -> _Path:
        """Path to the user's config dir."""
        return _Path(super().user_config_dir)

    @property
    def user_data_dir(self) -> _Path:
        """Path to the user's data dir."""
        return _Path(super().user_data_dir)

    @property
    def user_cache_dir(self) -> _Path:
        """Path to the user's cache dir."""
        return _Path(super().user_cache_dir)


class AppFiles(AppDirs):
    """Files for app to interact with."""

    @property
    def cache_file(self) -> _Path:
        """Path to the app's cache file."""
        return self.user_cache_dir / "files.json"

    @property
    def durations_file(self) -> _Path:
        """Path to the app's durations file."""
        return self.user_data_dir / "durations.json"

    @property
    def global_config_file(self) -> _Path:
        """Path to the app's global config file."""
        return self.user_config_dir / f"{NAME}.toml"

    @property
    def global_config_file_backup(self) -> _Path:
        """Path to the app's global config backup file."""
        return self.user_config_dir / f"{NAME}.toml.bak"

    @property
    def home_config_file(self) -> _Path:
        """Path to the app's home config file."""
        return self.user_home_dir / f".{NAME}rc"

    @property
    def project_config_file(self) -> _Path:
        """Path to the app's project config file."""
        return self.user_project_dir / f".{NAME}rc"

    @property
    def pyproject_toml(self) -> _Path:
        """Path to the project's pyproject.toml file."""
        return self.user_project_dir / "pyproject.toml"



"""
pyaud._cli
==========
"""
import inspect as _inspect
import json as _json
import sys as _sys
from argparse import SUPPRESS as _SUPPRESS
from argparse import ArgumentParser as _ArgumentParser

from . import plugins as _plugins
from ._locations import NAME as _NAME
from ._utils import colors as _colors
from ._version import __version__


class Parser(_ArgumentParser):
    """Inherited ``ArgumentParser`` object for package args.

    Assign positional argument to ``module`` or if there is a positional
    argument accompanying the ``modules`` argument assign it to
    ``positional``.

    If ``modules`` has been called run the ``_module_help`` method for
    extended documentation on each individual module that can be called.
    Documentation will be parsed from the plugin's docstring.

    If a valid positional argument has been called and ``SystemExit``
    has not been raised by the ``ArgumentParser`` help method then
    assign the chosen function to ``function`` to be called in ``main``.
    """

    def __init__(self) -> None:
        super().__init__(prog=_colors.cyan.get(_NAME))
        self._mapping = _plugins.mapping()
        self._registered = _plugins.registered()
        self._returncode = 0
        self._add_arguments()
        self._version_request()
        self.args = self.parse_args()
        if self.args.module == "modules":
            _sys.exit(self._module_help())

    def _add_arguments(self) -> None:
        self.add_argument(
            "module",
            metavar="MODULE",
            choices=[*self._registered, "modules"],
            help="choice of module: [modules] to list all",
        )
        self.add_argument(
            "-c",
            "--clean",
            action="store_true",
            help="clean unversioned files prior to any process",
        )
        self.add_argument(
            "-f",
            "--fix",
            action="store_true",
            help="suppress and fix all fixable issues",
        )
        self.add_argument(
            "-n",
            "--no-cache",
            action="store_true",
            help="disable file caching",
        )
        self.add_argument(
            "-s",
            "--suppress",
            action="store_true",
            help="continue without stopping for errors",
        )
        self.add_argument(
            "-t",
            "--timed",
            action="store_true",
            help="track the length of time for each plugin",
        )
        self.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=0,
            help="incrementally increase logging verbosity",
        )
        self.add_argument(
            "--rcfile",
            action="store",
            help="select file to override config hierarchy",
        )
        self.add_argument(
            "--version", action="store_true", help="show version and exit"
        )

        # pos argument following [modules] argument
        self.add_argument("pos", nargs="?", default=None, help=_SUPPRESS)

    def _print_module_docs(self) -> None:
        # iterate over ``modules`` object to print documentation on
        # particular module or all modules, depending on argument passed
        # to commandline
        print()
        for key in sorted(self._mapping):

            # keep a tab width of at least 1 space between key and
            # documentation
            # if all modules are printed adjust by the longest key
            tab = len(max(self._mapping, key=len)) + 1
            doc = _inspect.getdoc(self._mapping[key])
            if doc is not None:
                print(
                    "{}-- {}".format(
                        key.ljust(tab),
                        doc.splitlines()[0][:-1].replace("``", "`"),
                    )
                )

    def _print_module_summary(self) -> None:
        # print summary of module use if no module is selected or an
        # invalid module name os provided
        _colors.yellow.print(
            f"{_NAME} modules [<module> | all] for more on each module\n"
        )
        print(
            "modules = {}".format(
                _json.dumps(self._registered, indent=4, sort_keys=True)
            )
        )

    def _print_err(self) -> None:
        # announce selected module does not exist
        self._returncode = 0
        _colors.red.print(f"No such module: {self.args.pos}", file=_sys.stderr)

    def _module_help(self) -> int:
        self.print_usage()

        # module is selected or all has been entered
        if self.args.pos in (*self._mapping, "all"):

            # specific module has been entered for help output
            # filter the remaining modules from ``modules`` object
            if self.args.pos in self._mapping:
                positional = {self.args.pos: self._mapping[self.args.pos]}
                self._mapping.clear()
                self._mapping.update(positional)

            self._print_module_docs()

        # print module help summary if no argument to module has been
        # provided or argument has been provided but is not a valid
        # choice
        else:

            # argument has been provided which is not a valid option
            if self.args.pos is not None:
                self._print_err()

            self._print_module_summary()

        return self._returncode

    @staticmethod
    def _version_request() -> None:
        try:
            # the only exception for not providing positional args
            if _sys.argv[1] == "--version":
                print(__version__)
                _sys.exit(0)
        except IndexError:
            pass



"""
pyaud.config
============

Includes constants, functions, and singleton for Toml config parsing.

``toml`` can be used with the external API for retrieving parsed config.

    Configs are parsed in the following order:
        | ~/.config/pyaud/pyaud.toml
        | ~/.pyaudrc
        | .pyaudrc
        | pyproject.toml

The following methods can be called with ``toml``:

    .. code-block:: python

        toml.dump(
            self, fout: TextIO, obj: Optional[MutableMapping] = None
        ) -> str:

    Dump dict object to open file.

    If Optional[MutableMapping] is not provided, toml will use its
    own key-values.

    .. code-block:: python

        toml.dumps(self, obj: Optional[MutableMapping] = None) -> str

    Return dict object from open file as toml str.

    If Optional[MutableMapping] is not provided, toml will use its
    own key-values.

    .. code-block:: python

        toml.load(self, fin: TextIO, *args: Any) -> None

    Load dict object from open file.
"""
from __future__ import annotations

import copy as _copy
import importlib as _importlib
import inspect as _inspect
import logging as _logging
import logging.config as _logging_config
import os as _os
import shutil as _shutil
import typing as _t
from pathlib import Path as _Path

import appdirs as _appdirs
import tomli as _tomli
import tomli_w as _tomli_w

from . import _typing as _pt
from ._locations import NAME as _NAME
from ._locations import AppFiles as _AppFiles
from ._objects import MutableMapping as _MutableMapping

DEFAULT_CONFIG: _t.Dict[str, _t.Any] = dict(
    clean={"exclude": ["*.egg*", ".mypy_cache", ".env", "instance"]},
    logging={
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "standard": {
                "format": "%(asctime)s %(levelname)s %(name)s %(message)s"
            }
        },
        "handlers": {
            "default": {
                "class": "logging.handlers.TimedRotatingFileHandler",
                "formatter": "standard",
                "when": "d",
                "backupCount": 60,
                "filename": str(
                    _Path(_appdirs.user_log_dir(_NAME)) / f"{_NAME}.log"
                ),
            }
        },
        "root": {"level": "INFO", "handlers": ["default"], "propagate": False},
    },
    indexing={"exclude": ["whitelist.py", "conf.py", "setup.py"]},
    packages={"exclude": ["tests"]},
    audit={
        "modules": [
            "format",
            "format-docs",
            "format-str",
            "imports",
            "typecheck",
            "unused",
            "lint",
            "coverage",
            "readme",
            "docs",
        ]
    },
    addopts=["timed"],
)


class _Toml(_MutableMapping):
    """Base class for all ``toml`` object interaction."""

    def dumps(self, __obj: _t.Optional[_t.Dict[str, _t.Any]] = None) -> str:
        """Native ``dump(from)s(tr)`` method to include encoder.

        :param __obj: Mutable mapping dict-like object.
        :return: str object in toml encoded form.
        """
        return _tomli_w.dumps(dict(self) if __obj is None else dict(__obj))

    def loads(self, __s: str, *args: str) -> None:
        """Native ``load (from file)`` method.

        :param __s: Toml as str.
        :param args: Keys to search for.
        """
        obj = _tomli.loads(__s)
        for arg in args:
            obj = obj.get(arg, obj)

        self.update(obj)


class TempEnvVar:
    """Temporarily set a mutable mapping key-value pair.

    Set key-value whilst working within the context manager. If key
    already exists then change the key back to its original value. If
    key does not already exist then delete it so the environment is
    returned to its original state.

    :param obj: Mutable mapping to temporarily change.
    :param kwargs: Key-values to temporarily change in supplied object.
    """

    def __init__(self, obj: _t.MutableMapping, **kwargs: str) -> None:
        self._obj = obj
        self._default = {k: obj.get(k) for k in kwargs}
        self._obj.update(kwargs)

    def __enter__(self) -> TempEnvVar:
        return self

    def __exit__(
        self, exc_type: _pt.Exc, exc_val: _pt.ExcVal, exc_tb: _pt.ExcTB
    ) -> None:
        for key, value in self._default.items():
            if value is None:
                try:
                    del self._obj[key]
                except KeyError:

                    # in the case that key gets deleted within context
                    pass
            else:
                self._obj[key] = self._default[key]


def configure_global(app_files: _AppFiles) -> None:
    """Setup object with default config settings.

    Create config file with default config settings if one does not
    already exist.

    Load base config file which may, or may not, still have the default
    settings configured.

    :param app_files: App file locations object.
    """
    default_config = _copy.deepcopy(DEFAULT_CONFIG)
    toml.update(default_config)
    if app_files.global_config_file.is_file():
        while True:
            try:
                toml.loads(app_files.global_config_file.read_text())
                _shutil.copyfile(
                    app_files.global_config_file,
                    app_files.global_config_file_backup,
                )
                break

            except _tomli.TOMLDecodeError:
                if app_files.global_config_file_backup.is_file():
                    _os.rename(
                        app_files.global_config_file_backup,
                        app_files.global_config_file,
                    )
                else:
                    break

    app_files.global_config_file.write_text(toml.dumps())


def load_config(
    app_files: _AppFiles, opt: _t.Optional[_t.Union[str, _os.PathLike]] = None
) -> None:
    """Load configs in order, each one overriding the previous.

    :param app_files: App file locations object.
    :param opt: Optional extra path which will override all others.
    """
    files = [
        app_files.global_config_file,
        app_files.home_config_file,
        app_files.project_config_file,
        app_files.pyproject_toml,
    ]
    if opt is not None:
        files.append(_Path(opt))

    for file in files:
        if file.is_file():
            toml.loads(file.read_text(), "tool", _NAME)


def _extract_logger(default: _t.Dict[str, _t.Any]) -> _logging.Logger:
    # return the logging object
    parts = default["class"].split(".")
    module = _importlib.import_module(".".join(parts[:-1]))
    return getattr(module, parts[-1])


def _filter_default(
    default: _t.Dict[str, _t.Any], logger: _logging.Logger
) -> _t.Dict[str, _t.Any]:
    # filter out any invalid kwargs for logging config

    # this will be invalid, so re-add after
    cls = default["class"]
    formatter = default["formatter"]

    # inspect logger's signature for kwargs that it can take
    filter_keys = [
        param.name
        for param in _inspect.signature(
            logger  # type: ignore
        ).parameters.values()
        if param.kind == param.POSITIONAL_OR_KEYWORD
    ]

    # delete all the keys from the default section that are not valid
    for key in dict(default):
        if key not in filter_keys:
            del default[key]

    # re-add the str representation of the configured class
    default.update({"class": cls, "formatter": formatter})

    return default


def configure_logging(verbose: int = 0) -> None:
    """Set loglevel.

    If ``-v`` flag passed to commandline decrease runtime loglevel for
    every repeat occurrence.

    ``-vvvv`` will always set logging to ``DEBUG``.

    Default loglevel is set in the toml config and overridden by
    environment variable if there is one.

    :param verbose: Level to raise log verbosity by.
    """
    levels = ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL")
    config = dict(toml["logging"])
    default = config["handlers"]["default"]
    filename = default.get("filename")

    # get the logger object to filter out any invalid kwargs
    logger = _extract_logger(default)
    _filter_default(default, logger)

    # create logging dir and it's parents if they do not exist already
    if filename is not None:
        _Path(filename).expanduser().parent.mkdir(exist_ok=True, parents=True)

    # tweak loglevel if commandline argument is provided
    config["root"]["level"] = levels[
        max(0, levels.index(config["root"]["level"]) - verbose)
    ]

    # load values to ``logging``
    _logging_config.dictConfig(config)


toml = _Toml()
configure_global(_AppFiles())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyaud']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'environs>=9.4.0,<10.0.0',
 'gitspy>=0,<1',
 'lsfiles>=0,<1',
 'object-colors>=2.0.1,<3.0.0',
 'pyaud-plugins>=0.8',
 'spall>=0,<1',
 'tomli-w>=1.0.0,<2.0.0',
 'tomli>=2.0.1,<3.0.0']

entry_points = \
{'console_scripts': ['pyaud = pyaud.__main__:main']}

setup_kwargs = {
    'name': 'pyaud',
    'version': '4.1.0',
    'description': 'Framework for writing Python package audits',
    'long_description': 'pyaud\n=====\n.. image:: https://img.shields.io/badge/License-MIT-yellow.svg\n    :target: https://opensource.org/licenses/MIT\n    :alt: License\n.. image:: https://img.shields.io/pypi/v/pyaud\n    :target: https://pypi.org/project/pyaud/\n    :alt: PyPI\n.. image:: https://github.com/jshwi/pyaud/actions/workflows/ci.yml/badge.svg\n    :target: https://github.com/jshwi/pyaud/actions/workflows/ci.yml\n    :alt: CI\n.. image:: https://results.pre-commit.ci/badge/github/jshwi/pyaud/master.svg\n   :target: https://results.pre-commit.ci/latest/github/jshwi/pyaud/master\n   :alt: pre-commit.ci status\n.. image:: https://github.com/jshwi/pyaud/actions/workflows/codeql-analysis.yml/badge.svg\n    :target: https://github.com/jshwi/pyaud/actions/workflows/codeql-analysis.yml\n    :alt: CodeQL\n.. image:: https://codecov.io/gh/jshwi/pyaud/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/jshwi/pyaud\n    :alt: codecov.io\n.. image:: https://readthedocs.org/projects/pyaud/badge/?version=latest\n    :target: https://pyaud.readthedocs.io/en/latest/?badge=latest\n    :alt: readthedocs.org\n.. image:: https://img.shields.io/badge/python-3.8-blue.svg\n    :target: https://www.python.org/downloads/release/python-380\n    :alt: python3.8\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n    :alt: Black\n.. image:: https://img.shields.io/badge/linting-pylint-yellowgreen\n    :target: https://github.com/PyCQA/pylint\n    :alt: pylint\n\nFramework for writing Python package audits\n-------------------------------------------\n\nThe ``pyaud`` framework is designed for writing modular audits for Python packages\n\nAudits can be run to fail, such as when using CI, or include a fix\n\nFixes can be written for whole directories or individual files\n\nPlugins can be written for manipulating files\n\nSupports single script plugins\n\nInstallation\n------------\n\nPyPi\n****\n\n``pip install pyaud``\n\nDevelopment\n***********\n\n``poetry install``\n\nUsage\n-----\n\n.. code-block:: console\n\n    usage: pyaud [-h] [-c] [-f] [-n] [-s] [-t] [-v] [--rcfile RCFILE] [--version] MODULE\n\n    positional arguments:\n      MODULE           choice of module: [modules] to list all\n\n    optional arguments:\n      -h, --help       show this help message and exit\n      -c, --clean      clean unversioned files prior to any process\n      -f, --fix        suppress and fix all fixable issues\n      -n, --no-cache   disable file caching\n      -s, --suppress   continue without stopping for errors\n      -t, --timed      track the length of time for each plugin\n      -v, --verbose    incrementally increase logging verbosity\n      --rcfile RCFILE  select file to override config hierarchy\n      --version        show version and exit\n\nPlugins\n-------\n\n``pyaud`` will search for a plugins package in the project root\n\nTo register a plugin package ensure it is importable and prefix the package with ``pyaud_``\n\nThe name ``pyaud_plugins`` is reserved and will be automatically imported\n\nTo view available plugins see ``pyaud-plugins`` `README <https://github.com/jshwi/pyaud-plugins/blob/master/README.rst>`_ or run ``pyaud modules all``\n\nFor writing plugins see `docs <https://jshwi.github.io/pyaud/pyaud.html#pyaud-plugins>`_\n\nConfigure\n---------\n\nConfiguration of settings can be made with the following toml syntax files (overriding in this order):\n\n    | ~/.config/pyaud/pyaud.toml\n    | ~/.pyaudrc\n    | .pyaudrc\n    | pyproject.toml\n\nA config can be generated with `pyaud generate-rcfile`\n\nPrefix each key with ``tool.pyaud`` when using pyproject.toml\n',
    'author': 'jshwi',
    'author_email': 'stephen@jshwisolutions.com',
    'maintainer': 'jshwi',
    'maintainer_email': 'stephen@jshwisolutions.com',
    'url': 'https://pypi.org/project/pyaud/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
pyaud._utils
============
"""
from __future__ import annotations

import typing as _t
from pathlib import Path as _Path

import setuptools as _setuptools
from gitspy import Git as _Git
from object_colors import Color as _Color

from . import _config

colors = _Color()
git = _Git()

colors.populate_colors()


def get_packages() -> _t.List[str]:
    """Return list of Python package names currently in project.

    Prevent dot separated subdirectories (import syntax) as args are
    evaluated by path.

    Only return the parent package's name.

    :raises PythonPackageNotFoundError: Raised if no package can be
        found.
    :return: List of Python packages.
    """
    packages = list(
        {
            i.split(".", maxsplit=1)[0]
            for i in _setuptools.find_packages(
                # in response to an update to `setuptools` stubs:
                # - error: Argument "where" has incompatible type
                #   "Path"; expected "str"
                where=str(_Path.cwd()),
                exclude=_config.toml["packages"]["exclude"],
            )
        }
    )
    return sorted(packages)


def package() -> _t.Optional[str]:
    """Return name of primary Python package.

    :raises PythonPackageNotFoundError: Raised if no primary package can
        be determined.
    :return: Name of primary Python package.
    """
    # at least one package will be returned or an error would have been
    # raised
    packages = get_packages()

    # if there is only one package then that is the default
    if len(packages) == 1:
        return packages.pop()

    # if there are multiple packages found then look for a configured
    # package name that matches one of the project's packages
    package_name = _config.toml["packages"].get("name")
    if package_name in packages:
        return package_name

    # if there are multiple packages found, and none of the above two
    # apply, then the package with the same name as the project root (if
    # it exists) is the default
    repo = _Path.cwd().name
    if repo in packages:
        return repo

    return None


def get_commit_hash() -> _t.Optional[str]:
    """Get the hash of the current commit.

    :return: A ``str`` containing the hash of the commit, or None if no
        hash can be provided.
    """
    git.rev_parse("HEAD", capture=True, suppress=True)
    try:
        return git.stdout()[0]
    except IndexError:
        return None


def working_tree_clean() -> bool:
    """Check if working tree clean.

    :return: Working tree clean? True or False.
    """
    git.stdout()  # [...] -> void; clear stdout, if it exists
    git.status("--short", capture=True)
    return not git.stdout()



"""
pyaud._cache
============
"""
from __future__ import annotations

import hashlib as _hashlib
import typing as _t
from pathlib import Path as _Path

from . import exceptions as _exceptions
from ._indexing import IndexedState as _IndexedState
from ._indexing import files as _files
from ._locations import AppFiles as _AppFiles
from ._objects import JSONIO as _JSONIO
from ._objects import BasePlugin as _BasePlugin
from ._utils import colors as _colors
from ._utils import get_commit_hash as _get_commit_hash
from ._utils import working_tree_clean as _working_tree_clean


class HashMapping(_JSONIO):
    """Persistent data object.

    :param project: Project name.
    :param cls: Audit that this class is running in.
    :param commit: Commit that this audit is being run on.
    """

    _FB = "fallback"

    def __init__(
        self,
        project: str,
        cls: _t.Type[_BasePlugin],
        commit: _t.Optional[str] = None,
    ) -> None:
        super().__init__()
        self._project = project
        self._commit = commit or self._FB
        self._cls = str(cls)
        self._session: _t.Dict[str, str] = {}

    def tag(self, tag: str) -> None:
        """Tag commit key with a prefix.

        :param tag: Prefix to tag commit key with.
        """
        self._commit = f"{tag}-{self._commit}"

    def match_file(self, path: _Path) -> bool:
        """Match selected class against a file relevant to it.

        :param path: Path to the file to check if it has changed.
        :return: Is the file a match (not changed)? True or False.
        """
        relpath = str(path.relative_to(_Path.cwd()))
        newhash = _hashlib.new(  # type: ignore
            "md5", path.read_bytes(), usedforsecurity=False
        ).hexdigest()
        return newhash == self._session.get(relpath)

    def save_hash(self, path: _Path) -> None:
        """Populate file hash.

        :param path: Path to hash.
        """
        relpath = str(path.relative_to(_Path.cwd()))
        if path.is_file():
            newhash = _hashlib.new(  # type: ignore
                "md5", path.read_bytes(), usedforsecurity=False
            ).hexdigest()
            self._session[relpath] = newhash
        else:
            if relpath in self._session:
                del self._session[relpath]

    def read(self, path: _Path) -> None:
        """Read from file to object.

        :param path: Path to cache file.
        """
        super().read(path)
        project = self.get(self._project, {})
        fallback = project.get(self._FB, {})
        project[self._commit] = project.get(self._commit, fallback)
        self._session = project[self._commit].get(self._cls, {})

    def write(self, path: _Path) -> None:
        """Write data to file.

        :param path: Path to cache file.
        """
        cls = {self._cls: dict(self._session)}
        self[self._project] = {self._FB: cls, self._commit: cls}
        super().write(path)


class FileCacher:  # pylint: disable=too-few-public-methods
    """Handle caching of file(s).

    :param cls: Audit that this class is running in.
    :param func: Call function belonging to cls.
    :param app_files: App file locations object.
    :param args: Args that can be passed from other plugins.
    :param kwargs: Boolean flags for subprocesses.
    """

    def __init__(
        self,
        cls: _t.Type[_BasePlugin],
        func: _t.Callable[..., int],
        app_files: _AppFiles,
        *args: str,
        **kwargs: bool,
    ) -> None:
        self._cls = cls
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.no_cache = self.kwargs.get("no_cache", False)
        self._cache_file_path = app_files.cache_file
        self.hashed = HashMapping(
            app_files.user_project_dir.name, self._cls, _get_commit_hash()
        )
        if not _working_tree_clean():
            self.hashed.tag("uncommitted")

        self.hashed.read(app_files.cache_file)

    def _on_completion(self, *paths: _Path) -> None:
        self._cls.logger().debug("writing to %s", self._cache_file_path)
        for path in paths:
            self.hashed.save_hash(path)

        self.hashed.write(self._cache_file_path)

    def _cache_files(self) -> int:
        returncode = 0
        with _IndexedState() as state:
            for file in list(_files):
                if self.hashed.match_file(file):
                    self._cls.logger().debug("hit: %s", file)
                    _files.remove(file)
                else:
                    self._cls.logger().debug("miss: %s", file)
                    if self._cls.cache_all:
                        state.restore()
                        break

            if not _files and state.length:
                _colors.green.bold.print(
                    "No changes have been made to audited files"
                )
            else:
                returncode = self.func(*self.args, **self.kwargs)

            if not returncode:
                self._on_completion(*_files)

        return returncode

    def _cache_file(self) -> int:
        returncode = 0
        file = self._cls.cache_file
        if file is not None:
            path = _Path.cwd() / file
            try:
                returncode = self.func(*self.args, **self.kwargs)
            except _exceptions.AuditError as err:
                self._on_completion(path)
                raise _exceptions.AuditError(str(err)) from err

            if (
                not returncode
                and path.is_file()
                and self.hashed.match_file(path)
            ):
                self._cls.logger().debug("hit: %s", path)
                _colors.green.print(
                    "No changes have been made to audited file"
                )
                return 0

            self._cls.logger().debug("miss: %s", path)
            self._on_completion(path)

        return returncode

    def files(
        self, func: _t.Callable[..., int], *args: str, **kwargs: bool
    ) -> int:
        """Wrap ``__call__`` with a hashing function.

        :param func: Function to wrap.
        :param args: Args that can be passed from other plugins.
        :param kwargs: Boolean flags for subprocesses.
        :return: Wrapped function.
        """
        no_cache = kwargs.get("no_cache", False)
        self._cls.logger().info(
            "NO_CACHE=%s, %s.cache=%s, %s.cache_file=%s",
            no_cache,
            self._cls.__name__,
            self._cls.cache,
            self._cls.__name__,
            self._cls.cache_file,
        )
        if not no_cache:
            if self._cls.cache_file is not None:
                return self._cache_file()

            if self._cls.cache:
                return self._cache_files()

        self._cls.logger().info("skipping reading and writing to disk")
        return func(*args, **kwargs)
